"""Display the UI for a thing on the screen"""

import random

import serge.common
import serge.engine
import serge.sound
import serge.events
import serge.blocks.actors
import serge.blocks.layout
import serge.blocks.visualblocks
import serge.actor
from theme import G, theme
import common


class ThingUI(serge.actor.MountableActor):
    """Represents a thing on the screen"""

    pic_name_prefix = 'pic'
    drops_gold = False
    drops_health = False

    relaxation_for_hp = 0.5
    relaxation_limit = 0.1
    dead_visibility = 0.0

    death_sound = None

    def __init__(self, name, tag, model_actor):
        """Initialise the thing"""
        super(ThingUI, self).__init__(name, tag)
        #
        self.is_alive = True
        self.model_actor = model_actor

    def addedToWorld(self, world):
        """Added to the world"""
        super(ThingUI, self).addedToWorld(world)
        #
        self.setLayerName('ui')
        #
        # Picture
        self.pic = serge.actor.Actor('%s picture' % self.name, 'pic')
        self.pic.setSpriteName('%s-%s' % (self.name.lower(), self.pic_name_prefix))
        self.pic.setLayerName('ui-back')
        self.mountActor(self.pic, G('%s-pic-posn' % self.tag))
        #
        # Name
        self.thing_name = serge.blocks.actors.StringText(
            '%s picture' % self.name, 'pic', G('%s-name' % self.tag),
            colour=G('%s-name-colour' % self.tag),
            font_name=G('%s-font-name' % self.tag),
            font_size=G('%s-font-size' % self.tag),
            justify='center',
        )
        self.thing_name.setLayerName('ui-front')
        self.mountActor(self.thing_name, G('%s-name-posn' % self.tag))
        #
        # Slots for armour etc
        self.description_slots = serge.blocks.layout.VerticalBar(
            'vbox', 'vbox',
            G('%s-vbox-width' % self.tag),
            G('%s-vbox-height' % self.tag),
        )
        self.mountActor(self.description_slots, G('%s-vbox-posn' % self.tag))
        self.description_slots.setLayerName('ui-front')
        #
        # Slots for different named items
        self.slots = []
        for i in range(3):
            item = serge.blocks.actors.FormattedText(
                '%s slot' % self.tag, 'slot',
                format='%(slot_name)s %(slot_value)s',
                colour=G('%s-slot-%i-colour' % (self.tag, i+1)),
                font_name=G('%s-slot-font-name' % self.tag),
                font_size=G('%s-slot-font-size' % self.tag),
                justify='center',
                slot_name='slot %s' % i,
                slot_value='00HP',
            )
            item.setLayerName('ui-front')
            self.description_slots.addActor(item, 'ui-front')
            self.slots.append(item)
        #
        # HP Display
        self.hp_display = serge.actor.Actor('%s hp' % self.tag, 'hp-bar')
        self.hp_display.visual = serge.blocks.visualblocks.ProgressBar(
            G('%s-hp-size' % self.tag),
            value_ranges=[(0.0, 100.0, G('hp-colour'))],
        )
        self.mountActor(self.hp_display, G('%s-hp-posn' % self.tag))
        self.hp_display.setLayerName('ui-front')
        self.hp_display.visual.value = 50.0
        #
        self.hp_text = serge.blocks.actors.FormattedText(
            '%s hp text' % self.tag, 'hp text',
            format='%(hp_current)d / %(hp_max)d',
            colour=G('hp-colour'),
            font_size=G('hp-font-size'),
            justify='center',
            hp_current=10.0,
            hp_max=15.0,
        )
        self.hp_text.setLayerName('ui-front')
        self.mountActor(self.hp_text, G('player-hp-text-posn'))
        #
        # Click Display
        self.click_display = serge.actor.Actor('%s hp' % self.tag, 'click-bar')
        self.click_display.visual = serge.blocks.visualblocks.VerticalProgressBar(
            G('%s-click-size' % self.tag),
            value_ranges=[(0.0, self.model_actor.auto_click_interval, G('click-colour'))],
        )
        self.mountActor(self.click_display, G('%s-click-posn' % self.tag))
        self.click_display.setLayerName('ui-front')
        self.click_display.visual.value = 0.0
        #
        self.click_text = serge.blocks.actors.FormattedText(
            '%s click text' % self.tag, 'click text',
            format='%(click_current)d / %(click_max)d',
            colour=G('click-colour'),
            font_size=G('click-font-size'),
            justify='center',
            click_current=0.0,
            click_max=self.model_actor.auto_click_interval,
        )
        self.click_text.setLayerName('ui-front')
        self.click_text.setAngle(90.0)
        self.mountActor(self.click_text, G('%s-click-text-posn' % self.tag))
        #
        self.dead_hide_items = [
            self.hp_display,
            self.hp_text,
            self.click_text,
            self.click_display,
            self.thing_name,
            self.description_slots,
        ]
        #
        # Set initial state
        self.updateModelDisplay()

    def updateModelDisplay(self):
        """Update the display if it has changed because of power ups"""
        #self.thing_name.visual.setText(self.model_actor.class_name)
        self.hp_text.setValue('hp_max', self.model_actor.max_hp)
        self.hp_text.setValue('hp_current', self.model_actor.hp)
        self.hp_display.visual._checkRanges([(0, self.model_actor.max_hp, G('hp-colour'))])
        self.hp_display.visual.value = self.model_actor.hp

    def getAlive(self):
        """Return the alive status"""
        return self.is_alive

    def setAlive(self, value):
        """Set the alive status"""
        self.is_alive = value
        if not self.is_alive:
            for item in self.dead_hide_items:
                item.visible = False
            if self.dead_visibility:
                self.pic.visual.setAlpha(self.dead_visibility)
            else:
                self.pic.visible = False

        else:
            for item in self.dead_hide_items:
                item.visible = True
            if self.dead_visibility:
                self.pic.visual.setAlpha(1.0)
            else:
                self.pic.visible = True

    def takeAction(self, strike):
        """Show that a strike occurred"""
        if strike.target == self.model_actor and self.is_alive:
            #
            colour_name = 'strike-%s' % strike.damage_type
            self.text_item = DamageIndicator(
                'damage', 'damage',
                format='%(amount)d %(source)s',
                colour=G(colour_name),
                font_size=G('strike-font-size'),
                justify='center',
                amount=strike.damage,
                source=strike.actor.class_name if strike.actor.class_name != 'Thing' else '',
            )
            #
            # Play sound for being hurt
            if strike.damage_type == 'enemy-dot':
                serge.sound.Sounds.play('hurt')
                self.text_item.setValue('source', 'Poison')
            #
            self.text_item.setLayerName('ui-damage')
            x = random.randrange(*G('%s-action-x-range' % self.tag))
            y = random.randrange(*G('%s-action-y-range' % self.tag))
            #
            self.mountActor(self.text_item, (x, y))

    def updateActor(self, interval, world):
        """Update the actor"""
        super(ThingUI, self).updateActor(interval, world)
        #
        # Gradually change the display of the HP
        current_hp = self.hp_display.visual.value
        new_hp = self.relaxation_for_hp * current_hp + (1.0 - self.relaxation_for_hp) * self.model_actor.hp
        if abs(new_hp - self.model_actor.hp) < self.relaxation_limit:
            new_hp = self.model_actor.hp
        #
        self.hp_display.visual.value = new_hp
        self.hp_text.setValue('hp_current', new_hp)
        #
        # Update attack timer
        self.click_display.visual.value = min(self.model_actor.auto_click_interval, self.model_actor.attack_timer)
        #
        # Check if we died
        if self.is_alive != self.model_actor.is_alive:
            self.setAlive(self.model_actor.is_alive)
            if self.death_sound:
                serge.sound.Sounds.play(self.death_sound)
            if self.drops_gold:
                gold_item = GoldIndicator(
                    'gold-item', 'gold-item',
                    format='$%(amount)d',
                    colour=G('gold-text-colour'),
                    font_size=G('gold-item-font-size'),
                    justify='center',
                    amount=self.model_actor.value * common.GOLD_MULTIPLIER,
                )
                gold_item.setLayerName('ui-gold')
                x = random.randrange(*G('%s-action-x-range' % self.tag))
                y = random.randrange(*G('%s-action-y-range' % self.tag))
                #
                self.mountActor(gold_item, (x, y))
                gold_item.linkEvent(serge.events.E_MOUSE_OVER, self.mouseOverGold)
            #
            if self.drops_health:
                health_item = HealthIndicator(
                    'health-item', 'health-item',
                    format='%(amount)dHP',
                    colour=G('health-text-colour'),
                    font_size=G('health-item-font-size'),
                    justify='center',
                    amount=int(random.randrange(1.0, int(2.0 + self.model_actor.max_hp / G('health-multiplier')))),
                )
                health_item.setLayerName('ui-gold')
                x = random.randrange(*G('%s-action-x-range' % self.tag))
                y = random.randrange(*G('%s-action-y-range' % self.tag))
                #
                self.mountActor(health_item, (x, y))
                health_item.linkEvent(serge.events.E_MOUSE_OVER, self.mouseOverHealth)

    def mouseOverGold(self, obj, arg):
        """Mouse over the gold"""
        self.log.info('Collecting gold')
        self.processEvent((common.E_COLLECT_GOLD, obj))

    def mouseOverHealth(self, obj, arg):
        """Mouse over the health"""
        self.log.info('Collecting health')
        self.processEvent((common.E_COLLECT_HEALTH, obj))


class PlayerUI(ThingUI):
    """Display the player"""

    dead_visibility = 0.5
    death_sound = 'player-death'

    def addedToWorld(self, world):
        """Added to the world"""
        super(PlayerUI, self).addedToWorld(world)
        #
        self.click_display.visible = False
        self.click_text.visible = False
        self.slots[0].visible = False
        #
        self.updateModelDisplay()
        #
        self.thing_name.value = 'Fin'

    def updateModelDisplay(self):
        """Update the display"""
        super(PlayerUI, self).updateModelDisplay()
        #
        text_value = 'ATK%d' % self.model_actor.click_damage
        self.slots[2].visual.setText(text_value)
        text_value = 'AC%d' % self.model_actor.ac
        self.slots[1].visual.setText(text_value)


class SideKickUI(ThingUI):
    """Display the sidekick"""

    pic_name_prefix = 'side'

    def addedToWorld(self, world):
        """Added to the world"""
        super(SideKickUI, self).addedToWorld(world)
        #
        self.hp_display.visible = False
        self.hp_text.visible = False
        #
        self.slots[0].active = False
        self.slots[1].active = False
        self.slots[2].active = False
        #
        self.dead_hide_items = [
            self.click_text,
            self.click_display,
            self.thing_name,
            self.description_slots,
        ]
        #
        self.thing_name.value = {
            'sidekick-a': 'April',
            'sidekick-b': 'George',
            'sidekick-c': 'Nova',
        }[self.name]

    def updateModelDisplay(self):
        """Update the display"""
        super(SideKickUI, self).updateModelDisplay()
        #
        text_value = '%s for ATK%s' % (self.model_actor.weapon_name, self.model_actor.click_damage)
        self.slots[2].visual.setText(text_value)
        self.slots[0].visible = False
        self.slots[1].visible = False
        self.slots[2].visible = False

    def takeAction(self, strike):
        """Take an action"""
        super(SideKickUI, self).takeAction(strike)
        #
        if strike.actor == self.model_actor and strike.action:
            sound = self.model_actor.weapon_name.replace(' ', '-').lower()
            self.log.info('Trying to play sound for %s, "%s"' % (strike.actor, sound))
            serge.sound.Sounds.play(sound)


class AnimalUI(ThingUI):
    """Display the animal"""

    pic_name_prefix = 'animal'
    drops_gold = True
    drops_health = True
    death_sound = 'enemy-death'

    def addedToWorld(self, world):
        """Added to the world"""
        super(AnimalUI, self).addedToWorld(world)
        #
        self.slots[0].visible = False
        self.slots[1].visible = False
        self.slots[2].visible = True
        #
        text_value = 'ATK%d' % self.model_actor.click_damage
        self.slots[2].visual.setText(text_value)
        #
        self.pic.linkEvent(serge.events.E_LEFT_CLICK, self.clickedOn)

        zoom_size = 0.4 * (self.model_actor.click_damage / 26.0) + 0.6
        self.pic.setZoom(zoom_size)
        #
        self.thing_name.value = self.model_actor.real_name

        if self.model_actor.class_name.startswith('Shark-'):
            self.log.info('Boss shark detected')
            self.pic.setZoom(1.3)
            self.thing_name.value = G('boss-shark-names')[self.model_actor.class_name]

    def clickedOn(self, obj, arg):
        """Clicked on us"""
        self.log.info('Clicked on %s' % self.name)
        self.processEvent((common.E_ENEMY_CLICKED, self))

    def takeAction(self, strike):
        """Take an action"""
        super(AnimalUI, self).takeAction(strike)
        #
        if strike.actor == self.model_actor and strike.action:
            self.log.info('Trying to play sound for %s, "%s"' % (strike.actor, self.model_actor.class_name.lower()))
            serge.sound.Sounds.play(self.model_actor.class_name.lower())


class DamageIndicator(serge.blocks.actors.FormattedText):
    """Show damage and move upwards"""

    def __init__(self, *args, **kw):
        """Intialise"""
        super(DamageIndicator, self).__init__(*args, **kw)
        #
        self.lifetime = random.randrange(*G('strike-lifetime'))
        self.y_vel = random.randrange(*G('strike-y-vel'))

    def updateActor(self, interval, world):
        """Move the actor"""
        self.moveTo(self.x, self.y + interval / 1000.0 * self.y_vel, override_lock=True)
        self.lifetime -= interval
        if self.lifetime <= 0.0:
            world.scheduleActorRemoval(self)
        #
        self.visual.setAlpha(self.visual._alpha * G('strike-fade'))


class GoldIndicator(serge.blocks.actors.FormattedText):
    """Show gold and move downwards"""

    def __init__(self, *args, **kw):
        """Initialise"""
        super(GoldIndicator, self).__init__(*args, **kw)
        #
        self.x_vel = random.randrange(*G('gold-x-vel'))
        self.y_vel = random.randrange(*G('gold-y-vel'))
        self.y_limit = G('gold-y-limit')
        self.y_acc = G('gold-y-acc')
        self.mouse = serge.engine.CurrentEngine().getMouse()

    def updateActor(self, interval, world):
        """Move the actor"""
        if self.y <= self.y_limit:
            self.moveTo(
                self.x + interval / 1000.0 * self.x_vel,
                self.y + interval / 1000.0 * self.y_vel,
                override_lock=True
            )
            self.y_vel += interval / 1000.0 * self.y_acc
        #
        pos = self.mouse.getScreenPoint()
        if pos.isInside(self):
            self.processEvent((serge.events.E_MOUSE_OVER, self))


class HealthIndicator(serge.blocks.actors.FormattedText):
    """Show health and move downwards"""

    def __init__(self, *args, **kw):
        """Initialise"""
        super(HealthIndicator, self).__init__(*args, **kw)
        #
        self.x_vel = random.randrange(*G('health-x-vel'))
        self.y_vel = random.randrange(*G('health-y-vel'))
        self.y_limit = G('health-y-limit')
        self.y_acc = G('health-y-acc')
        self.mouse = serge.engine.CurrentEngine().getMouse()

    def updateActor(self, interval, world):
        """Move the actor"""
        if self.y <= self.y_limit:
            self.moveTo(
                self.x + interval / 1000.0 * self.x_vel,
                self.y + interval / 1000.0 * self.y_vel,
                override_lock=True
            )
            self.y_vel += interval / 1000.0 * self.y_acc
        #
        pos = self.mouse.getScreenPoint()
        if pos.isInside(self):
            self.processEvent((serge.events.E_MOUSE_OVER, self))
